from .ply import lex
from .tag import *
    
tokens = (
    "LBRACE",
    "RBRACE",
    "LBRACKET",
    "RBRACKET",
    "BOOL",
    "BYTE",
    "SHORT",
    "FLOAT",
    "DOUBLE",
    "LONG",
    "INTEGER",
    "STRING",
    "NAME",
    "TYPE",
    "COLON",
    "SEMICOLON",
    "COMMA",
)

t_ignore = ' \t\r'
t_ignore_COMMENT = r'\#.*'

def t_BOOL(t):
    r'true|false'
    t.value = Bool(t.value == "true")
    return t

def t_NAME(t):    
    r'[a-zA-Z0-9._+-]+(?=:)'
    return t

def t_STRING(t):
    r'\"[^\"\\]*(?:\\.[^\"\\]*)*\"'
    t.value = String(t.value[1:-1].replace(r'\"', '"').replace(r'\\', '\\'))
    return t

def t_BYTE(t):
    r'\-?[0-9]+b'
    t.value = Byte(t.value[:-1])
    return t

def t_SHORT(t):
    r'\-?[0-9]+s'
    t.value = Short(t.value[:-1])
    return t

def t_FLOAT(t):
    r'\-?[0-9]+(?:\.[0-9]+)?(?:[eE][\+\-]?[0-9]+)?f'
    t.value = Float(t.value[:-1])
    return t

def t_DOUBLE(t):
    r'\-?[0-9]+(?:\.[0-9]+)?(?:[eE][\+\-]?[0-9]+)?d'
    t.value = Double(t.value[:-1])
    return t

def t_LONG(t):
    r'\-?[0-9]+L'
    t.value = Long(t.value[:-1])
    return t

def t_INTEGER(t):
    r'\-?[0-9]+'
    t.value = Integer(t.value)
    return t

def t_TYPE(t):
    r'[BIL]+(?=;)'
    return t

def t_newline(t):
    r'\n'
    t.lexer.lineno += 1

def t_LBRACE(t):
    r'\{'
    t.lexer.level += 1
    return t

def t_RBRACE(t):
    r'\}'
    t.lexer.level -= 1
    return t

def t_LBRACKET(t):
    r'\['
    t.lexer.level += 1
    return t

def t_RBRACKET(t):
    r'\]'
    t.lexer.level -= 1
    return t

def t_COLON(t):
    r'\:'
    return t

def t_SEMICOLON(t):
    r'\;'
    return t

def t_COMMA(t):
    r'\,'
    return t

def t_error(t):
    print(f"Illegal character {t.value[0]!r}")
    t.lexer.skip(1)

lexer = lex.lex()
lexer.level = 0


from .ply import yacc
from .token import tokens
from .tag import *

__all__ = ["get_parser", "loads", "load"]

def p_snbt(p):
    """snbt : compound"""
    p[0] = p[1]

def p_compound(p):
    """compound : LBRACE key_value_pairs RBRACE
                | LBRACE RBRACE"""
    if len(p) == 3:
        p[0] = Compound()
    else:
        p[0] = Compound(p[2])

def p_key_value_pairs(p):
    """key_value_pairs : key_value_pairs key_value_pair
                       | key_value_pairs COMMA key_value_pair
                       | key_value_pair"""
    if len(p) == 2:
        p[0] = [p[1]]
    elif len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = p[1] + [p[2]]

def p_key_value_pair(p):
    """key_value_pair : NAME COLON value
                      | STRING COLON value"""
    p[0] = (p[1], p[3])

def p_value(p):
    """value : compound
                | list
                | array
                | BOOL
                | BYTE
                | SHORT
                | FLOAT
                | DOUBLE
                | LONG
                | INTEGER
                | STRING"""
    p[0] = p[1]

def p_list(p):
    """list : LBRACKET values RBRACKET
                | LBRACKET RBRACKET"""
    if len(p) == 3:
        p[0] = List()
    else:
        p[0] = List(p[2])

def p_array(p):
    """array : LBRACKET TYPE SEMICOLON values RBRACKET
                | LBRACKET TYPE SEMICOLON RBRACKET"""
    if len(p) == 5:
        p[0] = Array(p[2], [])
    else:
        p[0] = Array(p[2], p[4])

def p_values(p):
    """values : values value
              | values COMMA value
              | value"""
    if len(p) == 2:
        p[0] = [p[1]]
    elif len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = p[1] + [p[2]]
    
def p_error(p):
    raise ValueError(f"Syntax error at '{p.value}', line {p.lineno}, position {p.lexpos}, type {p.type}")

def get_parser():
    """Get the FTB snbt parser object.
    
    The parser object is ``yacc`` parser object that can be used to parse FTB snbt strings.
    
    Example
    -------
        >>> import ftb_snbt_lib as slib
        >>> parser = slib.get_parser()
        >>> snbt = parser.parse('''
            {
                some_tag: "some_value"
                another_tag: 1b
            }
            ''')
        >>> print(snbt)
        Compound({'some_tag': String('some_value'), 'another_tag': Byte(1)})
    """
    return yacc.yacc()

def loads(s: str) -> Compound:
    """Load an FTB snbt string and return the result as a Compound object.
    
    The loaded SNBT object consists of custom types: ``Compound``, ``List``, ``Array``, ``Numeric``, ``Bool``, and ``String``.
    
    When modifying the loaded object, use the provided custom types to ensure that the object is valid.

    Args
    ----
        s (str): The FTB snbt string to parse.

    Returns
    -------
        Compound: The loaded FTB snbt Compound object.
        
    Example
    -------
        >>> import ftb_snbt_lib as slib
        >>> snbt = slib.loads('''
            {
                some_tag: "some_value"
                another_tag: 1b
            }
            ''')
        >>> print(snbt)
        Compound({'some_tag': String('some_value'), 'another_tag': Byte(1)})
    """
    parser = get_parser()
    return parser.parse(s)

def load(fp) -> Compound:
    """    
    Load an FTB snbt file and return the result as a Compound object.
    
    The loaded SNBT object consists of custom types: ``Compound``, ``List``, ``Array``, ``Numeric``, ``Bool``, and ``String``.
    
    When modifying the loaded object, use the provided custom types to ensure that the object is valid.

    Args
    ----
        fp (``io.TextIOWrapper``): The FTB snbt file to parse (opened in reading text mode with a ``UTF-8`` encoding.)

    Returns
    -------
        Compound: The loaded FTB snbt Compound object.
        
    Example
    -------
        >>> import ftb_snbt_lib as slib
        >>> snbt = slib.load(open("path/to/file/some_file.snbt", "r", encoding="utf-8"))
        >>> type(snbt)
        <class 'ftb_snbt_lib.tag.Compound'>
    """
    parser = get_parser()
    return parser.parse(fp.read())


__all__ = ["Base", "Numeric", "NumericInteger", "Byte", "Short", "Integer", "Float", "Long", "Double", "Bool", "String", "List", "Array", "Compound"]

class InvalidTypeError(ValueError):
    def __init__(self, item, cls):
        self.item = item
        self.cls = cls
        super().__init__(f"{self.item!r} is not a valid type for the items of {cls.__name__}.\nThis error is likely caused by a type mismatch in a list.")

class CastError(ValueError):
    def __init__(self, item, subtype):
        self.item = item
        self.subtype = subtype
        super().__init__(f"{self.item!r} cannot be casted to {self.subtype.__name__}")

class Base:
    __slots__ = ()
    
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)
    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"

class Numeric(Base):
    __slots__ = ()
    write_func = "numeric"
    suffix = ""

class NumericInteger(Numeric, int):
    __slots__ = ()
    
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        limit = 2 ** (8 * cls.size - 1)
        cls.range = range(-limit, limit)
    
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        if int(self) not in cls.range:
            raise ValueError(f"{cls.__name__} value {int(self)} is out of range")
        return self

class Byte(NumericInteger):
    __slots__ = ()
    size = 1
    suffix = "b"

class Short(NumericInteger):
    __slots__ = ()
    size = 2
    suffix = "s"

class Integer(NumericInteger):
    __slots__ = ()
    size = 4

class Long(NumericInteger):
    __slots__ = ()
    size = 8
    suffix = "L"

class Float(Numeric, float):
    __slots__ = ()
    suffix = "f"

class Double(Numeric, float):
    __slots__ = ()
    suffix = "d"

class Bool(Base, int):
    __slots__ = ()
    write_func = "bool"
    
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        return self

class String(Base, str):
    __slots__ = ()
    write_func = "string"
    
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        return self

class List(Base, list):
    __slots__ = ()
    write_func = "list"
    subtype = None
    
    def __new__(cls, iterable=()):
        if cls.subtype is None:
            iterable = tuple(iterable)
            subtype = cls.infer_type(iterable)
            cls = cls[subtype]
        return super().__new__(cls, iterable)
    
    def __init__(self, iterable=()):
        super().__init__(map(self.cast_item, iterable))
    
    def __class_getitem__(cls, subtype):
        if subtype is None:
            return List
        return type(f"{cls.__name__}[{subtype.__name__}]", (List,), {"__slots__": (), "subtype": subtype})

    @staticmethod
    def infer_type(iterable):
        subtype = None
        
        for item in iterable:
            itemtype = type(item)
            if not issubclass(itemtype, Base):
                continue

            if subtype is None:
                subtype = itemtype
                if not issubclass(subtype, List):
                    return subtype
            elif subtype is not itemtype:
                sub_t, item_t = subtype, itemtype
                while issubclass(sub_t, List) and issubclass(item_t, List):
                    sub_t, item_t = sub_t.subtype, item_t.subtype

                if sub_t is None:
                    subtype = itemtype
                elif item_t is not None:
                    return None
            
        return subtype
    
    def __setitem__(self, index, value):
        if isinstance(index, int):
            super().__setitem__(index, self.cast_item(value))
        elif isinstance(index, slice):
            super().__setitem__(index, map(self.cast_item, value))
        else:
            raise TypeError(f"List indices must be integers or slices, not {index.__class__.__name__}")
        
    def append(self, value):
        super().append(self.cast_item(value))
    
    def extend(self, iterable):
        super().extend(map(self.cast_item, iterable))
    
    def insert(self, index, value):
        super().insert(index, self.cast_item(value))
    
    @classmethod
    def cast_item(cls, item):
        if cls.subtype is None:
            # If subtype cannot be inferred, accept Base types as-is
            if isinstance(item, Base):
                return item
            raise InvalidTypeError(item, cls)
        
        if not isinstance(item, cls.subtype):
            try:
                return cls.subtype(item)
            except Exception as e:
                raise CastError(item, cls.subtype) from e
            
        return item

class Array(Base, list):
    __slots__ = ()
    write_func = "array"
    subtype = None
    array_prefix = None

    def __new__(cls, subtype, iterable=()):
        if cls.subtype is None:
            cls = cls[subtype]
        return super().__new__(cls, subtype, iterable)
    
    def __init__(self, subtype, iterable=()):
        super().__init__(map(self.cast_item, iterable))

    def __class_getitem__(cls, subtype):
        if subtype == "B":
            return ByteArray
        if subtype == "I":
            return IntArray
        if subtype == "L":
            return LongArray
        raise TypeError(f"Array type must be specified with 'B', 'I', or 'L'")
    
    def __setitem__(self, index, value):
        if isinstance(index, int):
            super().__setitem__(index, self.cast_item(value))
        elif isinstance(index, slice):
            super().__setitem__(index, map(self.cast_item, value))
        else:
            raise TypeError(f"Array indices must be integers or slices, not {index.__class__.__name__}")
    
    def append(self, value):
        super().append(self.cast_item(value))
    
    def extend(self, iterable):
        super().extend(map(self.cast_item, iterable))
    
    def insert(self, index, value):
        super().insert(index, self.cast_item(value))

    @classmethod
    def cast_item(cls, item):
        if cls.subtype is None:
            raise InvalidTypeError(item, cls)
        
        if not isinstance(item, cls.subtype):
            try:
                return cls.subtype(item)
            except Exception as e:
                raise CastError(item, cls.subtype) from e
            
        return item

class ByteArray(Array):
    __slots__ = ()
    subtype = Byte
    array_prefix = "B"

class IntArray(Array):
    __slots__ = ()
    subtype = Integer
    array_prefix = "I"

class LongArray(Array):
    __slots__ = ()
    subtype = Long
    array_prefix = "L"
    
class Compound(Base, dict):
    __slots__ = ()
    write_func = "compound"


from contextlib import contextmanager
from .tag import *

__all__ = ["get_writer", "dumps", "dump"]

ESCAPE_SUBS = {
    "\\": "\\\\",
    '"': '\\"'
}

class Writer:
    def __init__(self, comma_sep: bool = False):
        self.indentation = "\t"
        self.indent = ""
        self.prev_indent = ""
        
        self.separator = "\n"
        if comma_sep:
            self.separator = "," + self.separator
        
    @contextmanager
    def depth(self):
        if self.indentation is None:
            yield
        else:
            prev = self.prev_indent
            self.prev_indent = self.indent
            self.indent += self.indentation
            yield
            self.indent = self.prev_indent
            self.prev_indent = prev
            
    def should_expand(self, tag: List | Array | Compound) -> bool:
        return (
            self.indentation is not None
            and (
                not self.prev_indent
                or isinstance(tag, (List, Array, Compound))
            )
        )

    def expand(self, separator: str, fmt: str) -> tuple[str, str]:
        return (
            f"{separator}{self.indent}",
            fmt.replace("{}", f"\n{self.indent}{{}}\n{self.prev_indent}"),
        )

    def write(self, tag: Base | str) -> str:
        func = getattr(self, f"write_{tag.write_func}", None)
        return func(tag) if func else f"{tag}"
    
    def write_numeric(self, tag: Numeric) -> str:
        func = int.__repr__ if isinstance(tag, int) else float.__repr__
        return f"{func(tag)}{tag.suffix}"
    
    def write_bool(self, tag: Bool) -> str:
        return "true" if tag else "false"
    
    def write_string(self, tag: String) -> str:
        for match, seq in ESCAPE_SUBS.items():
            tag = tag.replace(match, seq)
        return f'"{tag}"'
    
    def write_list(self, tag: List) -> str:
        separator, fmt = self.separator, "[{}]"

        if len(tag) == 0:
            return fmt.format(" ")
        if len(tag) == 1:
            return fmt.format(self.write(tag[0]))
        
        with self.depth():
            if self.should_expand(tag):
                separator, fmt = self.expand(separator, fmt)

            return fmt.format(separator.join(map(self.write, tag)))
        
    def write_array(self, tag: Array) -> str:
        separator, fmt = self.separator, f"[{tag.array_prefix};{{}}]"

        if len(tag) == 0:
            return fmt.format(" ")
        if len(tag) == 1:
            return fmt.format(self.write(tag[0]))
        
        with self.depth():
            if self.should_expand(tag):
                separator, fmt = self.expand(separator, fmt)

            return fmt.format(separator.join(map(self.write, tag)))
    
    def write_compound(self, tag: Compound) -> str:
        separator, fmt = self.separator, "{{{}}}"

        if len(tag) == 0:
            return fmt.format(" ")
        
        with self.depth():
            if self.should_expand(tag):
                separator, fmt = self.expand(separator, fmt)

            return fmt.format(
                separator.join(
                    f"{self.stringify_compound_key(key)}: {self.write(value)}"
                    for key, value in tag.items()
                )
            )
            
    def stringify_compound_key(self, key: String | str) -> str:
        if isinstance(key, String):
            return self.write_string(key)
        return key
    
def get_writer(comma_sep: bool = False) -> Writer:
    """Get the FTB snbt writer object.
    
    Args
    ----
        comma_sep (bool): Whether to separate tags with commas.
        
    Returns
    -------
        Writer: The FTB snbt writer object.
    
    Example
    -------
        >>> import ftb_snbt_lib as slib
        >>> writer = slib.get_writer()
        >>> snbt = slib.Compound({
        ...     'some_tag': slib.String('some_value'),
        ...     'another_tag': slib.Byte(1)
        ... })
        >>> dumped_snbt = writer.write(snbt)
        >>> print(dumped_snbt)
        {
            some_tag: "some_value"
            another_tag: 1b
        }
    """
    return Writer(comma_sep=comma_sep)

def dumps(tag: Base, comma_sep: bool = False) -> str:
    """Dump an FTB snbt tag to a string.
    
    A tag can be a ``Compound``, ``List``, ``Numeric``, ``Bool``, or ``String`` object.
    
    The returned string can be written to a file or sent over a network.

    Args
    ----
        tag (Base): The FTB snbt tag to dump.
        comma_sep (bool): Whether to separate tags with commas.

    Returns
    -------
        str: The dumped FTB snbt string.
        
    Example
    -------
        >>> import ftb_snbt_lib as slib
        >>> snbt = slib.Compound({
        ...     'some_tag': slib.String('some_value'),
        ...     'another_tag': slib.Byte(1)
        ... })
        >>> dumped_snbt = slib.dumps(snbt)
        >>> print(dumped_snbt)
        {
            some_tag: "some_value"
            another_tag: 1b
        }
    """
    return get_writer(comma_sep=comma_sep).write(tag) + "\n"

def dump(tag: Base, fp, comma_sep: bool = False) -> None:
    """Dump an FTB snbt tag to a file.

    Args
    ----
        tag (Base): The FTB snbt tag to dump.
        fp (io.TextIOWrapper): The file to write the FTB snbt tag to (opened in writing text mode with a UTF-8 encoding.)
        comma_sep (bool): Whether to separate tags with commas.
    
    Example
    -------
        >>> import ftb_snbt_lib as slib
        >>> snbt = slib.Compound({
        ...     'some_tag': slib.String('some_value'),
        ...     'another_tag': slib.Byte(1)
        ... })
        >>> slib.dump(snbt, open('some_file.snbt', 'w', encoding='utf-8'))
    """
    fp.write(dumps(tag, comma_sep=comma_sep))